;===============================================
; MouseTest_9x11_726.asm
;
; 
;
; Microchip Assembly program
; May 1, 2011
; David D. Kaufman
; TheVaporTrail.com
;===============================================


		;---------------------------------------------
		;	Specify Processor and include file
		;---------------------------------------------
		processor 16f726

#include "P16F726.INC"

		;---------------------------------------------
		;	Adjust warnings
		;---------------------------------------------
		errorlevel -302
		
		;---------------------------------------------
		;	Decimal
		;---------------------------------------------
		radix	dec
				
		;---------------------------------------------
		;	Chip configuration
		;---------------------------------------------
		;	Oscillator: Internal oscillator
		;	Lines are I/O (that is, not oscillator)
		;	Watchdog Timer: Off
		;	Power-up timer: On
		;---------------------------------------------
		__config _CONFIG1, _INTOSCIO & _WDT_OFF & _PWRT_EN


		;---------------------------------------------
		;	Constants
		;---------------------------------------------
#define		MULTIPLEX_DELAY		20

#define		CTR_XY				0x45	; Center pixel
#define		QUIET_THRESHOLD		 2		; Threshold for testing if mouse is still connected
#define		SCALING				 8		; Amount the X & Y values are divided by to reduce the movement
#define		CURSOR_X_MAX		10
#define		CURSOR_Y_MAX		 8
#define		CURSOR_ON_COUNT		10
#define		CURSOR_OFF_COUNT	20


		;	Flags
#define		FLAG_CURSOR_SHOWN	0
#define		FLAG_LEFT_BTN_PREV	1
#define		FLAG_RIGHT_BTN_PREV	2
#define		FLAG_ERASE_PIXEL	3
#define		FLAG_LEFT_BTN		4
#define		FLAG_RIGHT_BTN		5
#define		FLAG_XY_CHANGE		6
#define		FLAG_LEFT_BTN_CHNG	7

		;---------------------------------------------
		;	Build-config Constants
		;
		;	These change exactly what statements are
		;	assembled.
		;---------------------------------------------
#define		REVERSE_X_DIRECTION  0		; Change to '1' to reverse the X mouse direction (i.e., dX = -dX)
#define		REVERSE_Y_DIRECTION  0		; Change to '1' to reverse the Y mouse direciton (i.e., dY = -dY)

#define		SHOW_RAW_DATA		 0		; Change to '1' to show raw data on the display and disable clearing display

		;---------------------------------------------
		;	PS/2 Mouse Constants
		;
		;---------------------------------------------

		; PS2 Ports and Pins
#define		PS2_CLK_PORT		PORTB			; Clock line port, config, and pin
#define		PS2_CLK_TRIS		TRISB
#define		PS2_CLK_PIN			4
#define		PS2_DAT_PORT		PORTB			; Data line port, config, and pin
#define		PS2_DAT_TRIS		TRISB
#define		PS2_DAT_PIN			5

		; Flag Bits in PS2_FLAG
#define		PS2_BIT_ERROR		0				; Set if a read, write, or timeout error occurs
#define		PS2_BIT_PARITY		1				; Used to calculate parity during write
#define		PS2_BIT_NEW_PKT		2				; Set if a new packet is received in PS2_AcquirePacket

		; PS2 Commands
#define		PS2_CMD_EN_DAT_RPT		0xF4		; Enable Data Reporting

		; PS2 Errors.							; For all non-zero values, PS_FLAG_ERROR bit of PS_FLAGS will also be set.
#define		PS2_ERR_NONE		0				; No error during last call. 
#define		PS2_ERR_WRITE_FAIL	1				; Write failed
#define		PS2_ERR_READ_FAIL	2				; Read failed

		; PS2 Configuration
#define		PS2_LEVEL_CHANGE_TIMEOUT	0xff	; Number of times to loop waiting for the clock or data level to change
#define		PS2_NEW_PKT_TIMEOUT			5		; Number of times to call PS_WaitClockLow to see if a new packet is being sent by device
#define 	PS2_CLOCK_LOW_COUNT 		25		; Count for number of times to loop holding the clock line low in PS2_WriteByte. 
												; (Total time should be at least 100 usec)



		;---------------------------------------------
		;	Memory Variables
		;---------------------------------------------
			cblock	0x20
			COUNTER						; Delay counter variables
			TMP_CNT
			TEMP
			TEMP2
			FRAME:18					; Frame storage
			FRAMECOUNT					; Countdown to show current frame

			FLAGS						; See FLAGS_xxx, above, for bit usage

			CURSOR_X					; Cursor position
			CURSOR_X_FRACT
			CURSOR_Y
			CURSOR_Y_FRACT
			CURSOR_X_PREV				; Previous cursor position
			CURSOR_Y_PREV
			CURSOR_CNTR					; Cursor blink counter

			QUIET_CNT_LO				; Count of times that there was no packet from the mouse
			QUIET_CNT_HI				; (high byte)
			QUIET_FLAGS					; Used to track if we cleared the display and stage in animation

			;---------------------------------------------
			;	PS/2 Variables
			;---------------------------------------------
			PS2_FLAGS					; See PS2_FLAG_xxx #defines, above
			PS2_DATA					; Data, used for read and write
			PS2_COUNT					; Counter for PS2 routines
			PS2_RESULT					; Result, see PS2_ERR_xxx #defines, above
			PS2_STEP					; Debug variable; increments in read routine 
			PS2_DEV_FLAGS				; First byte of PS2 packet: flags and high bits of DX and DY
			PS2_DEV_DX					; Second byte: X movement ("delta X")
			PS2_DEV_DY					; Third byte: Y movement ("delta Y")


			endc

			cblock	0xA0
			FRAME_ALT:18				; Frame data backup; copied from/to FRAME
			endc

		;---------------------------------------------
		;	Vectors
		;---------------------------------------------
			org 0x00
			goto Start					; Power-on Vector



		;---------------------------------------------
		;	Main Program
		;---------------------------------------------
			org	0x05
Start		nop

			;-------------------------------  Configure the Ports
			bsf		STATUS, RP0
			clrf	TRISA						; All pins on PortA are output
			clrf	TRISB						; All pins on PortB are output
			clrf	TRISC						; All pins on PortC are output

			; Configure PS2 Pins
			bsf		PS2_CLK_TRIS, PS2_CLK_PIN	; ..change any specific pins to input
			bsf		PS2_DAT_TRIS, PS2_DAT_PIN	; 

			bcf		OPTION_REG, NOT_RBPU		; Global weak pull-up enable
			bsf		WPUB, PS2_CLK_PIN			; Turn-on weak pull-ups for clock and data pins on PORTB
			bsf		WPUB, PS2_DAT_PIN

			banksel	ANSELB
			clrf	ANSELB 						; Port B, when inputs, are digital
			banksel	ANSELA
			clrf	ANSELA 						; Port A, when inputs, are digital

			bcf		STATUS, RP0
			bcf		STATUS, RP1

			clrf	PORTA
			clrf	PORTB
			clrf	PORTC

			;-------------------------------  Initialize all variables
			clrf	FLAGS

			clrf	PS2_DEV_FLAGS				; Clear the PS2 variables
			clrf	PS2_DEV_DX
			clrf	PS2_DEV_DY

			clrf	QUIET_CNT_LO				; Clear the 'no activity' count
			clrf	QUIET_CNT_HI

			movlw	5							; Initialize the cursor position..
			movwf	CURSOR_X					; ..to the center of the 9x11 display
			movlw	4
			movwf	CURSOR_Y

			clrf	CURSOR_X_FRACT				; Clear the fractional part of the..
			clrf	CURSOR_Y_FRACT				; ..cursor position vars (used for scaling)

			;-------------------------------  Clear the display 

			call	ShowStartAnimation			; Show the start-up animation

			call	ClearFrame					; Clear the frame and..
			call	CopyFrameToFrameAlt			; ..the frame copy
			
			movlw	PS2_CMD_EN_DAT_RPT			; Tell the mouse to..
			call	PS2_WriteByte				; ..enable data reporting
			call	PS2_ReadByte				; Read the response, but ignore it

			;btfsc	PS2_FLAGS, PS2_BIT_ERROR	; If there was an error talking to the mouse, ..
			;goto	MouseFail					; ..then jump to the 'mouse failed' section

			call	PS2_DisableDevComm			; Disable device communication 

			btfsc	PS2_FLAGS, PS2_BIT_ERROR	; Did we get an error trynig to reach the mouse earlier?
			goto	MouseCheck					; Yes. Go to the mouse check section
			nop									; No. Continue to main loop

			bsf		FLAGS, FLAG_CURSOR_SHOWN	; Set-up the cursor blink
			movlw	CURSOR_ON_COUNT
			movwf	CURSOR_CNTR

MainLoop

			;-------------------------------   Display frame with cursor
			btfsc	FLAGS, FLAG_CURSOR_SHOWN	; Put the cursor on the display, but only while during..
			call	InvertCursorPixel			; ..the display section (we take it off after the display call)

			movlw	0x20						; Display the current frame for a short while
			call	DisplayFrame

			btfsc	FLAGS, FLAG_CURSOR_SHOWN	; Then turn off the cursor. Note that the FLAG_CURSOR_SHOWN is..
			call	InvertCursorPixel			; ..toggled as a way to blink the cursor

			;-------------------------------   Handle cursor blink
			decfsz	CURSOR_CNTR,F				; Decrement the counter that is used to control the blink rate..
			goto	$+7							; ..of the cursor. If the counter reaches zero, then we toggle..
			movlw	1<<FLAG_CURSOR_SHOWN		; ..the FLAG_CURSOR_SHOWN bit and reload the counter.
			xorwf	FLAGS,F						; 
			movlw	CURSOR_ON_COUNT				; Note that we use different count down values for 'cursor on'..
			btfss	FLAGS, FLAG_CURSOR_SHOWN	; ..and 'cursor off'. This allows us to show the difference..
			movlw	CURSOR_OFF_COUNT			; ..between when the cursor is over an 'on' pixel or an 'off' pixel
			movwf	CURSOR_CNTR
			nop

			;-------------------------------   Attempt packet acquire
			call	PS2_AcquirePacket			; Attempt to acquire a new packet from the mouse
			btfss	PS2_FLAGS, PS2_BIT_NEW_PKT	; Was a packet available?
			goto	NoActivity					; No packet; jump to 'no activity' section
			
			;-------------------------------   Show raw data on display (currently commented-out)
			if SHOW_RAW_DATA					; NOTE: This is commented-out
			movf	PS2_DEV_FLAGS,W				; Put the packet value directly on the display
			movwf	FRAME+0						; 
			movf	PS2_DEV_DX,W
			movwf	FRAME+2
			movf	PS2_DEV_DY,W
			movwf	FRAME+4
			movf	PS2_STEP,W
			movwf	FRAME+6
			endif

			;-------------------------------   Restart "no activity" counter and clear some flags
			clrf	QUIET_CNT_LO				; Restart "no activity" counter 
			clrf	QUIET_CNT_HI

			bcf		FLAGS, FLAG_XY_CHANGE		; Clear the position changed and..
			bcf		FLAGS, FLAG_LEFT_BTN_CHNG	; ..left button changed flags

			;-------------------------------   Save current values
			movf	CURSOR_X,W					; CURSOR_X_PREV = CURSOR_X				
			movwf	CURSOR_X_PREV
			movf	CURSOR_Y,W					; CURSOR_Y_PREV = CURSOR_Y
			movwf	CURSOR_Y_PREV

			bcf		FLAGS, FLAG_LEFT_BTN_PREV	; LeftBtnPrev = LeftBtn
			btfsc	FLAGS, FLAG_LEFT_BTN
			bsf		FLAGS, FLAG_LEFT_BTN_PREV
			bcf		FLAGS, FLAG_RIGHT_BTN_PREV	; RightBtnPrev = RightBtn
			btfsc	FLAGS, FLAG_RIGHT_BTN
			bsf		FLAGS, FLAG_RIGHT_BTN_PREV

			;-------------------------------   Move button flags from packet to FLAGS var (for convenience)
			bcf		FLAGS, FLAG_LEFT_BTN		; LeftBtn = mouse left btn
			btfsc	PS2_DEV_FLAGS, 0
			bsf		FLAGS, FLAG_LEFT_BTN
			bcf		FLAGS, FLAG_RIGHT_BTN		; RightBtn = mouse right btn
			btfsc	PS2_DEV_FLAGS, 1
			bsf		FLAGS, FLAG_RIGHT_BTN

			;-------------------------------   ACTION: Left Button pressed --> Clear display
			if !SHOW_RAW_DATA
			btfsc	FLAGS, FLAG_RIGHT_BTN		; If the left button is pressed..
			call	ClearFrame					; ..then clear the display
			endif

			;-------------------------------   ACTION: Both Buttons pressed --> Load a pattern
			if !SHOW_RAW_DATA
			movf	FLAGS,W										; If both buttons are pressed..
			andlw	(1<<FLAG_LEFT_BTN) | (1<<FLAG_RIGHT_BTN)	
			xorlw	(1<<FLAG_LEFT_BTN) | (1<<FLAG_RIGHT_BTN)
			btfsc	STATUS,Z
			call	LoadStar									; ..then load a pattern
			nop
			endif

			;-------------------------------   Update the X cursor position and keep value in range
												; Note that this code is doing 'scaling' of the data
												; that is receives from the mouse. The mouse actually
												; support this in hardware, and we should simply be
												; sending the 'scale command'. I can't really think of
												; any benefit to doing this in software.

			movf	PS2_DEV_DX,W				; Update the X position
			if REVERSE_X_DIRECTION
			sublw	0							; Reverse the direction (W = -W)
			endif
			addwf	CURSOR_X_FRACT,F
X_T_Lp		btfss	CURSOR_X_FRACT,7			; Is it negative?
			goto	X_P_Lp						; No. Go to positive loop
X_N_Lp		decf	CURSOR_X,F					; Yes. Decrement X and..
			movlw	SCALING						; ..add the scaling factor to..
			addwf	CURSOR_X_FRACT,F			; ..the fractional part
			goto	X_T_Lp
X_P_Lp		movlw	SCALING						; Is _FRACT >= SCALING?
			subwf	CURSOR_X_FRACT, W			; ..ie, _FRACT - W >= 0
			btfsc	STATUS, C
			goto	X_T_Ex						; No. Exit loop
			incf	CURSOR_X,F
			movlw	-SCALING
			addwf	CURSOR_X_FRACT,F
			goto	X_P_Lp

X_T_Ex		btfsc	CURSOR_X,7					; Keep X value in range (0..10)
			clrf	CURSOR_X
			movlw	CURSOR_X_MAX
			subwf	CURSOR_X,W
			movlw	CURSOR_X_MAX
			btfsc	STATUS,C
			movwf	CURSOR_X

			;-------------------------------   Update the Y cursor position and keep value in range
			movf	PS2_DEV_DY,W				; Update the Y position
			if REVERSE_Y_DIRECTION
			sublw	0							; Reverse the direction (W = -W)
			endif
			addwf	CURSOR_Y_FRACT,F
Y_T_Lp		btfss	CURSOR_Y_FRACT,7			; Is it negative?
			goto	Y_P_Lp						; No. Go to positive loop
Y_N_Lp		decf	CURSOR_Y,F					; Yes. Decrement Y and..
			movlw	SCALING						; ..add the scaling factor to..
			addwf	CURSOR_Y_FRACT,F			; ..the fractional part
			goto	Y_T_Lp
Y_P_Lp		movlw	SCALING						; Is _FRACT >= SCALING?
			subwf	CURSOR_Y_FRACT, W
			btfsc	STATUS, C
			goto	Y_T_Ex						; No. Exit loop
			incf	CURSOR_Y,F
			movlw	-SCALING
			addwf	CURSOR_Y_FRACT,F
			goto	Y_P_Lp

Y_T_Ex		btfsc	CURSOR_Y,7					; Keep Y value in range (0..8)
			clrf	CURSOR_Y
			movlw	CURSOR_Y_MAX
			subwf	CURSOR_Y,W
			movlw	CURSOR_Y_MAX
			btfsc	STATUS,C
			movwf	CURSOR_Y

			;-------------------------------   Check for change in X, Y, or Left Button value
			movf	CURSOR_X_PREV,W				; Did it change from the previous XY value?
			subwf	CURSOR_X,W
			btfss	STATUS,Z
			bsf		FLAGS, FLAG_XY_CHANGE
			movf	CURSOR_Y_PREV,W
			subwf	CURSOR_Y,W
			btfss	STATUS,Z
			bsf		FLAGS, FLAG_XY_CHANGE
			nop

																	; Button change testing
			movlw	(1<<FLAG_LEFT_BTN) | (1<<FLAG_LEFT_BTN_PREV)	; Mask of current and previous button states
			andwf	FLAGS,W											; Get only these two bits into W
			btfsc	STATUS,Z										; If zero, then both prev. and curr. are 'up'
			goto	XYB_Change_Test									; Skip ahead
			xorlw	(1<<FLAG_LEFT_BTN) | (1<<FLAG_LEFT_BTN_PREV)	; Invert both bits
			btfsc	STATUS,Z										; If zero, then both prev. and curr. are 'down'
			goto	XYB_Change_Test									; Skip ahead
			bsf		FLAGS, FLAG_LEFT_BTN_CHNG						; Set flag saying button changed
			
			;-------------------------------   Did XY or (left) button change?
XYB_Change_Test	
			movlw	(1<<FLAG_LEFT_BTN_CHNG) | (1<<FLAG_XY_CHANGE)	; Mask for only the change flags
			andwf	FLAGS,W
			btfsc	STATUS,Z										; If non-zero, then we have at least one change
			goto	XYB_NoChange									; Nothing changed; so go to end of loop
			
			;-------------------------------   Something changed
			movlw	(1<<FLAG_LEFT_BTN_CHNG) | (1<<FLAG_XY_CHANGE)	; If either the position or the left button..
			andwf	FLAGS,W											; ..changed..
			btfsc	STATUS,Z										;
			goto	$+3												;
			btfsc	FLAGS, FLAG_LEFT_BTN							; ..and if the left button is pressed..
			call	InvertCursorPixel								; ..invert the pixel at the cursor
			nop

			btfss	FLAGS, FLAG_XY_CHANGE							; If the position changed, then restart..
			goto	$+4												; ..the cursor flag timer
			movlw	CURSOR_ON_COUNT
			movwf	CURSOR_CNTR
			bsf		FLAGS, FLAG_CURSOR_SHOWN
			nop


XYB_NoChange
			goto	MainLoop_End

			;-------------------------------  "No activity" count-down section
NoActivity
			incf	QUIET_CNT_LO,F				; Increment the two-byte "no activity" counter
			btfss	STATUS,Z
			goto	MainLoop_End
			incf	QUIET_CNT_HI,F
			movlw	QUIET_THRESHOLD				; If the counter reaches our threshold..
			subwf	QUIET_CNT_HI,W				; ..then we jump to a section that will..
			btfsc	STATUS,Z					; ..check to see if the mouse is still..
			goto	MouseCheck					; ..connected.

			;-------------------------------  End of the main loop
MainLoop_End
			goto	MainLoop					; Go back to top of loop

			;-------------------------------  Mouse check section
MouseCheck
			clrf	QUIET_FLAGS					; Clear the flags for this section
MouseCheckLoop
			call	PS2_EnableDevComm			; Allow the mouse to freely send events
			movlw	PS2_CMD_EN_DAT_RPT			; Send the 'enable data reporting' command
			call	PS2_WriteByte
			btfss	PS2_FLAGS, PS2_BIT_ERROR	; Did is succeed?
			goto	$+4							; Yes. Goto read the response
			btfss	QUIET_FLAGS,0				; No. First time thru?
			call	CopyFrameToFrameAlt			; Copy to backup
			goto	MC_Animation				; No. Failed to write the byte; enter an animation phase
			call	PS2_ReadByte				; Yes. Read the response
			btfsc	PS2_FLAGS, PS2_BIT_ERROR	; Did is succeed?
			goto	MC_Animation				; No. Failed to read the response; enter an animation phase
			call	PS2_DisableDevComm			; Yes. Go back to suppressing events from the mouse
			clrf	QUIET_CNT_LO				; Reset the "no activity" counter
			clrf	QUIET_CNT_HI
			btfsc	QUIET_FLAGS, 0				; Did we clear the diaplay and put something else on it?
			call	CopyFrameAltToFrame			; Yes, so clear it again to show we have the mouse back.
			goto	MainLoop					; Go back to the main loop

MC_Animation
			bsf		QUIET_FLAGS,0				; Set a flag to show that we cleared the display
			call	ClearFrame					; And clear it
			movlw	CTR_XY						; Draw either a "forward" or "backward" slash..
			call	SetPixel					; ..depending on bit [1:1] of QUIET_FLAGS
			movlw	CTR_XY + 0x10 - 0x01
			btfss	QUIET_FLAGS,1
			movlw	CTR_XY + 0x10 + 0x01
			call	SetPixel
			movlw	CTR_XY - 0x10 + 0x01
			btfss	QUIET_FLAGS,1
			movlw	CTR_XY - 0x10 - 0x01
			call	SetPixel
			movlw	2							; Toggle bit [1:1] of QUIET_FLAGS..
			xorwf	QUIET_FLAGS,F				; ..so that other slash is drawn next time through

			movlw	0x08						; Show the current frame for a little while,..
			movwf	TMP_CNT						; ..both as a way to do a delay before checking..
			movlw	0xff						; ..for the mouse again and because we need to..
			call	DisplayFrame				; ..show the "searching for mouse" pattern
			decfsz	TMP_CNT,F
			goto	$-3

			goto	MouseCheckLoop				; Check again for the mouse

			;-------------------------------  Mouse Fail section (not used)
MouseFail
			call	ClearFrame					; Clear the frame
			movlw	CTR_XY						; Draw an "x"..
			call	SetPixel					; ..one..
			movlw	CTR_XY + 0x10 + 0x01		; ..pixel..
			call	SetPixel					; ..at..
			movlw	CTR_XY + 0x10 - 0x01		; ..a..
			call	SetPixel					; ..time..
			movlw	CTR_XY - 0x10 + 0x01
			call	SetPixel
			movlw	CTR_XY - 0x10 - 0x01
			call	SetPixel

			movlw	0x80						; Show the frame..
			call	DisplayFrame				; ..and..
			goto	$-2							; ..loop forever


			;-------------------------------  End of "application"; subroutines follow


ShowStartAnimation
			call	ClearFrame
			movlw	CTR_XY
			call	SetPixel
			movlw	CTR_XY + 0x10
			call	SetPixel
			movlw	CTR_XY - 0x10
			call	SetPixel
			movlw	CTR_XY + 0x01
			call	SetPixel
			movlw	CTR_XY - 0x01
			call	SetPixel
			movlw	CTR_XY + 0x20
			call	SetPixel
			movlw	CTR_XY - 0x20
			call	SetPixel
			movlw	CTR_XY + 0x02
			call	SetPixel
			movlw	CTR_XY - 0x02
			call	SetPixel
			movlw	CTR_XY + 0x30
			call	SetPixel
			movlw	CTR_XY - 0x30
			call	SetPixel
			movlw	CTR_XY + 0x03
			call	SetPixel
			movlw	CTR_XY - 0x03
			call	SetPixel

			movlw	0x03
			movwf	TMP_CNT
			movlw	0xff
			call	DisplayFrame
			decfsz	TMP_CNT,F
			goto	$-3

			call	ClearFrame
			movlw	CTR_XY
			call	SetPixel
			movlw	CTR_XY + 0x10
			call	SetPixel
			movlw	CTR_XY - 0x10
			call	SetPixel
			movlw	CTR_XY + 0x01
			call	SetPixel
			movlw	CTR_XY - 0x01
			call	SetPixel
			movlw	CTR_XY + 0x20
			call	SetPixel
			movlw	CTR_XY - 0x20
			call	SetPixel
			movlw	CTR_XY + 0x02
			call	SetPixel
			movlw	CTR_XY - 0x02
			call	SetPixel

			movlw	0x03
			movwf	TMP_CNT
			movlw	0xff
			call	DisplayFrame
			decfsz	TMP_CNT,F
			goto	$-3

			call	ClearFrame
			movlw	CTR_XY
			call	SetPixel
			movlw	CTR_XY + 0x10
			call	SetPixel
			movlw	CTR_XY - 0x10
			call	SetPixel
			movlw	CTR_XY + 0x01
			call	SetPixel
			movlw	CTR_XY - 0x01
			call	SetPixel

			movlw	0x03
			movwf	TMP_CNT
			movlw	0xff
			call	DisplayFrame
			decfsz	TMP_CNT,F
			goto	$-3

			call	ClearFrame
			movlw	CTR_XY
			call	SetPixel

			movlw	0x03
			movwf	TMP_CNT
			movlw	0xff
			call	DisplayFrame
			decfsz	TMP_CNT,F
			goto	$-3

			call	ClearFrame

			return

;============================================================================================
;=====
;=====		Sub-frame edit routines
;=====
;=====
;============================================================================================

		;---------------------------------------------
		;	PutCursorInW
		;
		;---------------------------------------------
PutCursorInW
			movf	CURSOR_X,W
			andlw	0x0f
			movwf	TEMP
			
			swapf	CURSOR_Y,W
			andlw	0xf0
			iorwf	TEMP,W
			return

		;---------------------------------------------
		;	SetCursorPixel
		;
		;---------------------------------------------
SetCursorPixel
			call	PutCursorInW
			goto	SetPixel

		;---------------------------------------------
		;	ClearCursorPixel
		;
		;---------------------------------------------
ClearCursorPixel
			call	PutCursorInW
			goto	ClearPixel

		;---------------------------------------------
		;	XorCursorPixel
		;
		;---------------------------------------------
InvertCursorPixel
			call	PutCursorInW
			goto	InvertPixel

		;---------------------------------------------
		;	GetCursorPixel
		;
		;---------------------------------------------
GetCursorPixel
			call	PutCursorInW
			goto	GetPixel

		;---------------------------------------------
		;	SetPixel
		;
		;	High half into W, 
		;	Low half into W
		;---------------------------------------------
SetPixel
			call	ComputePixelInfo
			iorwf	INDF,F
			return

		;---------------------------------------------
		;	SetPixel
		;
		;---------------------------------------------
ClearPixel
			call	ComputePixelInfo
			xorlw	0xff
			andwf	INDF,F
			return

		;---------------------------------------------
		;	XorPixel
		;
		;---------------------------------------------
InvertPixel
			call	ComputePixelInfo
			xorwf	INDF,F
			return

		;---------------------------------------------
		;	GetPixel
		;
		;	STATUS[Z]: contains result
		;---------------------------------------------
GetPixel
			call	ComputePixelInfo
			andwf	INDF,W
			return

		;---------------------------------------------
		;	ComputePixelInfo
		;
		;	High half into W, 
		;	Low half into W
		;---------------------------------------------
ComputePixelInfo
			movwf	TEMP2

			; Compute address of first FRAME byte to update by
			; adding the constant FRAME and twice the value in xxx.
			; Put this in the indexed register
			swapf	TEMP2,W
			andlw	0x0f
			movwf	TEMP
			addwf	TEMP,F
			movlw	FRAME
			addwf	TEMP,W
			movwf	FSR						; ..and store

			movlw	0x0f
			andwf	TEMP2,f
			btfss	TEMP2,3					; Is it >8?
			goto	$+3
			bcf		TEMP2,3
			incf	FSR,F
			nop

			bcf		STATUS,C				; Create a one-bit mask
			movlw	1
			movwf	TEMP
			movf	TEMP2,F
			btfsc	STATUS,Z
			goto	$+4
			rlf		TEMP,F
			decfsz	TEMP2,F
			goto	$-2
			movf	TEMP,W

			return


;============================================================================================
;=====
;=====		Misc full-frame routines
;=====
;============================================================================================

		;---------------------------------------------
		;	Copy Frame to Diaplay (FRAME_ALT)
		;
		;---------------------------------------------
CopyFrameToFrameAlt
			movf	FRAME+0,W
			bsf		STATUS, RP0
			movwf	FRAME_ALT+0
			bcf		STATUS, RP0

			movf	FRAME+1,W
			bsf		STATUS, RP0
			movwf	FRAME_ALT+1
			bcf		STATUS, RP0

			movf	FRAME+2,W
			bsf		STATUS, RP0
			movwf	FRAME_ALT+2
			bcf		STATUS, RP0

			movf	FRAME+3,W
			bsf		STATUS, RP0
			movwf	FRAME_ALT+3
			bcf		STATUS, RP0

			movf	FRAME+4,W
			bsf		STATUS, RP0
			movwf	FRAME_ALT+4
			bcf		STATUS, RP0

			movf	FRAME+5,W
			bsf		STATUS, RP0
			movwf	FRAME_ALT+5
			bcf		STATUS, RP0

			movf	FRAME+6,W
			bsf		STATUS, RP0
			movwf	FRAME_ALT+6
			bcf		STATUS, RP0

			movf	FRAME+7,W
			bsf		STATUS, RP0
			movwf	FRAME_ALT+7
			bcf		STATUS, RP0

			movf	FRAME+8,W
			bsf		STATUS, RP0
			movwf	FRAME_ALT+8
			bcf		STATUS, RP0

			movf	FRAME+9,W
			bsf		STATUS, RP0
			movwf	FRAME_ALT+9
			bcf		STATUS, RP0

			movf	FRAME+10,W
			bsf		STATUS, RP0
			movwf	FRAME_ALT+10
			bcf		STATUS, RP0

			movf	FRAME+11,W
			bsf		STATUS, RP0
			movwf	FRAME_ALT+11
			bcf		STATUS, RP0

			movf	FRAME+12,W
			bsf		STATUS, RP0
			movwf	FRAME_ALT+12
			bcf		STATUS, RP0

			movf	FRAME+13,W
			bsf		STATUS, RP0
			movwf	FRAME_ALT+13
			bcf		STATUS, RP0

			movf	FRAME+14,W
			bsf		STATUS, RP0
			movwf	FRAME_ALT+14
			bcf		STATUS, RP0

			movf	FRAME+15,W
			bsf		STATUS, RP0
			movwf	FRAME_ALT+15
			bcf		STATUS, RP0

			movf	FRAME+16,W
			bsf		STATUS, RP0
			movwf	FRAME_ALT+16
			bcf		STATUS, RP0

			movf	FRAME+17,W
			bsf		STATUS, RP0
			movwf	FRAME_ALT+17
			bcf		STATUS, RP0
			return

		;---------------------------------------------
		;	Copy Frame to Diaplay (FRAME_ALT)
		;
		;---------------------------------------------
CopyFrameAltToFrame
			bsf		STATUS, RP0
			movf	FRAME_ALT+0,W
			bcf		STATUS, RP0
			movwf	FRAME+0

			bsf		STATUS, RP0
			movf	FRAME_ALT+1,W
			bcf		STATUS, RP0
			movwf	FRAME+1

			bsf		STATUS, RP0
			movf	FRAME_ALT+2,W
			bcf		STATUS, RP0
			movwf	FRAME+2

			bsf		STATUS, RP0
			movf	FRAME_ALT+3,W
			bcf		STATUS, RP0
			movwf	FRAME+3

			bsf		STATUS, RP0
			movf	FRAME_ALT+4,W
			bcf		STATUS, RP0
			movwf	FRAME+4

			bsf		STATUS, RP0
			movf	FRAME_ALT+5,W
			bcf		STATUS, RP0
			movwf	FRAME+5

			bsf		STATUS, RP0
			movf	FRAME_ALT+6,W
			bcf		STATUS, RP0
			movwf	FRAME+6

			bsf		STATUS, RP0
			movf	FRAME_ALT+7,W
			bcf		STATUS, RP0
			movwf	FRAME+7

			bsf		STATUS, RP0
			movf	FRAME_ALT+8,W
			bcf		STATUS, RP0
			movwf	FRAME+8

			bsf		STATUS, RP0
			movf	FRAME_ALT+9,W
			bcf		STATUS, RP0
			movwf	FRAME+9

			bsf		STATUS, RP0
			movf	FRAME_ALT+10,W
			bcf		STATUS, RP0
			movwf	FRAME+10

			bsf		STATUS, RP0
			movf	FRAME_ALT+11,W
			bcf		STATUS, RP0
			movwf	FRAME+11

			bsf		STATUS, RP0
			movf	FRAME_ALT+12,W
			bcf		STATUS, RP0
			movwf	FRAME+12

			bsf		STATUS, RP0
			movf	FRAME_ALT+13,W
			bcf		STATUS, RP0
			movwf	FRAME+13

			bsf		STATUS, RP0
			movf	FRAME_ALT+14,W
			bcf		STATUS, RP0
			movwf	FRAME+14

			bsf		STATUS, RP0
			movf	FRAME_ALT+15,W
			bcf		STATUS, RP0
			movwf	FRAME+15

			bsf		STATUS, RP0
			movf	FRAME_ALT+16,W
			bcf		STATUS, RP0
			movwf	FRAME+16

			bsf		STATUS, RP0
			movf	FRAME_ALT+17,W
			bcf		STATUS, RP0
			movwf	FRAME+17

			return

		;---------------------------------------------
		;	Clear Frame
		;	
		;---------------------------------------------
ClearFrame
			clrf	FRAME
			clrf	FRAME+2
			clrf	FRAME+4
			clrf	FRAME+6
			clrf	FRAME+8
			clrf	FRAME+10
			clrf	FRAME+12
			clrf	FRAME+14
			clrf	FRAME+16
			clrf	FRAME+1
			clrf	FRAME+3
			clrf	FRAME+5
			clrf	FRAME+7
			clrf	FRAME+9
			clrf	FRAME+11
			clrf	FRAME+13
			clrf	FRAME+15
			clrf	FRAME+17
			return

		;---------------------------------------------
		;	InvertFrame
		;	
		;---------------------------------------------
InvertFrame
			movlw	0xff
			xorwf	FRAME,f
			xorwf	FRAME+2,f
			xorwf	FRAME+4,f
			xorwf	FRAME+6,f
			xorwf	FRAME+8,f
			xorwf	FRAME+10,f
			xorwf	FRAME+12,f
			xorwf	FRAME+14,f
			xorwf	FRAME+16,f
			movlw	0x07
			xorwf	FRAME+1,f
			xorwf	FRAME+3,f
			xorwf	FRAME+5,f
			xorwf	FRAME+7,f
			xorwf	FRAME+9,f
			xorwf	FRAME+11,f
			xorwf	FRAME+13,f
			xorwf	FRAME+15,f
			xorwf	FRAME+17,f
			return


;============================================================================================
;=====
;=====		Misc Routines
;=====
;============================================================================================

		;---------------------------------------------
		;	Long Delay
		;	Max delay @ 4Mhz: 0.32768 sec
		;---------------------------------------------
LongDelay	movlw	0xff
AdjDelay	movwf	TEMP2
LoopLDA		movlw	0x00
			movwf	TEMP
LoopLDB		nop
			nop
			decfsz	TEMP,F
			goto	LoopLDB
			decfsz	TEMP2,F
			goto	LoopLDB
			return


;============================================================================================
;=====
;=====		Frame Display routines (non-interrupt)
;=====
;============================================================================================
		;---------------------------------------------
		;	ShowFrame
		;	
		;---------------------------------------------
ShowFrame
			movlw	0xff
			call	DisplayFrame
			movlw	0xff
			call	DisplayFrame
			movlw	0xff
			call	DisplayFrame
			movlw	0xff
			call	DisplayFrame
			movlw	0xff
			call	DisplayFrame
			movlw	0xff
			call	DisplayFrame
			return

		;---------------------------------------------
		;	ShowShortFrame
		;	
		;---------------------------------------------
ShowShortFrame
			movlw	0xff
			goto	DisplayFrame

		;---------------------------------------------
		;	ShowVeryShortFrame
		;	
		;---------------------------------------------
ShowVeryShortFrame
			movlw	0x40
			goto	DisplayFrame

		;---------------------------------------------
		;	DisplayFrame
		;	Display the current frame in FRAME storage
		;	for 'w' iterations 
		;---------------------------------------------
DisplayFrame
			movwf	FRAMECOUNT
LoopDF		call	DisplaySingleFrame
			decfsz	FRAMECOUNT,F
			goto	LoopDF
			return


		;---------------------------------------------
		;	DisplaySingleFrame
		;	Display the current frame in FRAME storage
		;	once.
		;	When exiting this routine leaves the 
		;	display disabled
		;---------------------------------------------
DisplaySingleFrame

			movf	PORTB,w		; Turn off column 9
			andlw	0xf7
			movwf	PORTB
			clrf	PORTC		; Insure columns 1 to 8 are off

			;===================== COLUMN 1 (of 1..9)
			movf	FRAME,w
			movwf	PORTA
			movf	FRAME+1,w
			andlw	0x7
			movwf	TEMP
			movf	PORTB,w
			andlw	0xf8
			iorwf	TEMP,w
			movwf	PORTB

			movlw	0x01
			movwf	PORTC
			call	MultiplexDelay
			clrf	PORTC

			;===================== COLUMN 2 (of 1..9)
			movf	FRAME+2,w
			movwf	PORTA
			movf	FRAME+3,w
			andlw	0x7
			movwf	TEMP
			movf	PORTB,w
			andlw	0xf8
			iorwf	TEMP,w
			movwf	PORTB

			movlw	0x02
			movwf	PORTC
			call	MultiplexDelay
			clrf	PORTC

			;===================== COLUMN 3 (of 1..9)
			movf	FRAME+4,w
			movwf	PORTA
			movf	FRAME+5,w
			andlw	0x7
			movwf	TEMP
			movf	PORTB,w
			andlw	0xf8
			iorwf	TEMP,w
			movwf	PORTB

			movlw	0x04
			movwf	PORTC
			call	MultiplexDelay
			clrf	PORTC

			;===================== COLUMN 4 (of 1..9)
			movf	FRAME+6,w
			movwf	PORTA
			movf	FRAME+7,w
			andlw	0x7
			movwf	TEMP
			movf	PORTB,w
			andlw	0xf8
			iorwf	TEMP,w
			movwf	PORTB

			movlw	0x08
			movwf	PORTC
			call	MultiplexDelay
			clrf	PORTC


			;===================== COLUMN 5 (of 1..9)
			movf	FRAME+8,w
			movwf	PORTA
			movf	FRAME+9,w
			andlw	0x7
			movwf	TEMP
			movf	PORTB,w
			andlw	0xf8
			iorwf	TEMP,w
			movwf	PORTB

			movlw	0x10
			movwf	PORTC
			call	MultiplexDelay
			clrf	PORTC


			;===================== COLUMN 6 (of 1..9)
			movf	FRAME+10,w
			movwf	PORTA
			movf	FRAME+11,w
			andlw	0x7
			movwf	TEMP
			movf	PORTB,w
			andlw	0xf8
			iorwf	TEMP,w
			movwf	PORTB

			movlw	0x20
			movwf	PORTC
			call	MultiplexDelay
			clrf	PORTC

			
			;===================== COLUMN 7 (of 1..9)
			movf	FRAME+12,w
			movwf	PORTA
			movf	FRAME+13,w
			andlw	0x7
			movwf	TEMP
			movf	PORTB,w
			andlw	0xf8
			iorwf	TEMP,w
			movwf	PORTB

			movlw	0x40
			movwf	PORTC
			call	MultiplexDelay
			clrf	PORTC

			
			;===================== COLUMN 8 (of 1..9)
			movf	FRAME+14,w
			movwf	PORTA
			movf	FRAME+15,w
			andlw	0x7
			movwf	TEMP
			movf	PORTB,w
			andlw	0xf8
			iorwf	TEMP,w
			movwf	PORTB
			movlw	0x80
			movwf	PORTC
			call	MultiplexDelay
			clrf	PORTC
			

			;===================== COLUMN 9 (of 1..9)
			movf	FRAME+16,w
			movwf	PORTA
			movf	FRAME+17,w
			andlw	0x7
			movwf	TEMP
			movf	PORTB,w
			andlw	0xf8
			iorwf	TEMP,w
			;movwf	PORTB

			;movf	PORTB,w			; Turn on column 9
			iorlw	0x08
			movwf	PORTB
			call	MultiplexDelay
			movf	PORTB,w			; Turn off..
			andlw	0xf0			; ..column 9 and rows 9..11
			movwf	PORTB

			clrf	PORTA			; Turn off rows 1..8

			return


		;---------------------------------------------
		;	Multiplex Delay
		;	Pre-set delay to be called between row
		;	updates on display
		;---------------------------------------------
MultiplexDelay	
			movlw	MULTIPLEX_DELAY
			movwf	COUNTER
LoopMD		nop
			decfsz	COUNTER,F
			goto	LoopMD
			return



;============================================================================================
;=====
;=====		PS/2 Routines
;=====
;=====		http://www.computer-engineering.org/ps2protocol/
;=====		http://www.computer-engineering.org/ps2mouse/
;=====
;============================================================================================


		;---------------------------------------------
		;	PS2: Disable Device Communications
		;
		;	This routine will set the clock line low.
		;	When the clock line is low, a PS/2 device
		;	will not attempt to send any packets.
		;
		;	This routine is used, along with 
		;	PS2_EnableDevComm, as a way to leave the
		;	mouse in streaming mode (the default) and
		;	then poll for changes.
		;	See PS2_AcquirePacket.
		;---------------------------------------------
PS2_DisableDevComm
			bcf		PS2_CLK_PORT, PS2_CLK_PIN	; Set the clock pin low
			banksel	PS2_CLK_TRIS
			bcf		PS2_CLK_TRIS, PS2_CLK_PIN	; Change pin to output 
			banksel	PS2_CLK_PORT
			return

		;---------------------------------------------
		;	PS2: Enable Device Communications
		;
		; 	See comments above, in PS2_DisableDevComm
		;---------------------------------------------
PS2_EnableDevComm
			banksel	PS2_CLK_TRIS
			bsf		PS2_CLK_TRIS, PS2_CLK_PIN	; Change pin to input 
			banksel	PS2_CLK_PORT
			return

		;---------------------------------------------
		;	PS2: Acquire Packet
		;
		;	This routine will release the hold on the
		;	clock line and wait a short time to see if
		;	the PS/2 device is sending a packet (by
		;	monitoring the clock line). If a packet
		;	is being sent, this routine calls 
		;	PS2_ReadPacket. If the read is successful, 
		;	the "new packet" bit is set in the 
		;	PS2_FLAGS variable.
		;---------------------------------------------
PS2_AcquirePacket
			bcf		PS2_FLAGS, PS2_BIT_NEW_PKT		; Clear the flag saying a packet is available

			call	PS2_EnableDevComm				; Release the clock line to allow device to send packet

			movlw	PS2_NEW_PKT_TIMEOUT				; Init the count for the number of times we wait
			movwf	PS2_COUNT						; ..for the clock to go low
			call	PS2_WaitClockLow				; Is the device starting to send a packet?
			btfss	PS2_FLAGS, PS2_BIT_ERROR		; Was there as error (e.g. timeout)?
			goto	$+4								; No. Line is low. Jump to 'read packet'
			decfsz	PS2_COUNT,F						; Timeout occurred. Decrement count and ..
			goto	$-4								; ..repeat if count has not yet ended
			goto	PS2_AP_Exit						; Clock did not go low; skip to end

			call	PS2_ReadPacket					; Attempt to read the packet
			btfss	PS2_FLAGS, PS2_BIT_ERROR		; Was an error returned?
			bsf		PS2_FLAGS, PS2_BIT_NEW_PKT		; No. Set the 'packet available' flag
PS2_AP_Exit
			call	PS2_DisableDevComm				; Pull the clock line high to prevent packets
			return

		;---------------------------------------------
		;	PS2: Read Packet
		;
		;	Reads three consecutive bytes from the PS/2
		;	device (currently assumed to be a mouse)
		;	and store the packets in the PS2_DEV_FLAGS,
		;	PS2_DEV_DX, and PS2_DEV_DY variables.
		;
		;	If an error occurs the PS2_BIT_ERROR flag 
		;	will be set and the PS2_RESULT will be
		;	non-zero
		;---------------------------------------------
PS2_ReadPacket
			call	PS2_ReadByte
			btfsc	PS2_FLAGS, PS2_BIT_ERROR
			goto	PS2_RP_Exit
			movf	PS2_DATA,W
			movwf	PS2_DEV_FLAGS

			call	PS2_ReadByte
			btfsc	PS2_FLAGS, PS2_BIT_ERROR
			goto	PS2_RP_Exit
			movf	PS2_DATA,W
			movwf	PS2_DEV_DX

			call	PS2_ReadByte
			btfsc	PS2_FLAGS, PS2_BIT_ERROR
			goto	PS2_RP_Exit
			movf	PS2_DATA,W
			movwf	PS2_DEV_DY

PS2_RP_Exit
			return

		;---------------------------------------------
		;	PS2: Write Byte
		;
		;	Input
		;	W: byte to write
		;
		;	Output
		;	PS_FLAGS[PS_BIT_ERROR]: Set if error, 
		;		cleared otherwise. If this is set, then
		;		result is set to a non-zero value
		;	PS_RESULT: result code. 
		;
		;	To write to a device, first bring the clock
		;	low for at least 100 usec, then bring the data
		;	low, and then release the clock line.
		;	The device will now drive the clock line and
		;	put out 11 pulses.
		;	For the first 8 pulses, the host will write
		;	the data bits, starting with the LSB.
		;	For the 9th pulse, the host will write the
		;	parity bit (ODD)
		;	The 10th pulse is a stop bit. The host must
		;	release the data line.
		;	The 11th pulse is an ACK from the device.
		;---------------------------------------------
PS2_WriteByte
			movwf	PS2_DATA						; Store byte to write
			movlw	PS2_ERR_NONE					; Set result to 'no error'
			movwf	PS2_RESULT
			bcf		PS2_FLAGS, PS2_BIT_ERROR		; Clear the error bit
			bsf		PS2_FLAGS, PS2_BIT_PARITY		; Initialize parity calculation variable. Parity is ODD

			bsf		PS2_CLK_PORT, PS2_CLK_PIN		; Start with lines high when switched to output
			nop
			bsf		PS2_DAT_PORT, PS2_DAT_PIN
			nop
			
			banksel	PS2_CLK_TRIS					; Set-up PS2 pins for output
			bcf		PS2_CLK_TRIS, PS2_CLK_PIN
			banksel	PS2_DAT_TRIS
			bcf		PS2_DAT_TRIS, PS2_DAT_PIN
			banksel	PS2_CLK_PORT

			;------ REQUEST WRITE ----
													; Bring the Clock line low for at least 100 microseconds.
			bcf		PS2_CLK_PORT, PS2_CLK_PIN		; Set clock line low
			movlw	PS2_CLOCK_LOW_COUNT				; Start a counter using the constant defined in the ..
			movwf	TEMP							; ..configuration section
			nop
			nop
			decfsz	TEMP,F
			goto	$-3
			
			bcf		PS2_DAT_PORT, PS2_DAT_PIN		; Bring the Data line low.
			nop
			nop

			bsf		PS2_CLK_PORT, PS2_CLK_PIN		; Release the Clock line
			banksel	PS2_CLK_TRIS					; Set-up PS2 Clock pin for input, because..
			bsf		PS2_CLK_TRIS, PS2_CLK_PIN		; ..the device drives the clock
			banksel	PS2_CLK_PORT
			
			movlw	8								; Init the number of bits to read
			movwf	PS2_COUNT

			;------ DATA BITS ------
PS2_WB_Loop		
			call	PS2_WaitClockLow				; Wait for the device to bring the Clock line low.
			btfsc	PS2_FLAGS, PS2_BIT_ERROR
			goto	PS2_WB_fail

			rrf		PS2_DATA,F						; Put the next bit on the Data line
			btfss	STATUS,C
			goto	$+2
			goto	$+3
			bcf		PS2_DAT_PORT, PS2_DAT_PIN
			goto	$+4
			bsf		PS2_DAT_PORT, PS2_DAT_PIN
			movlw	(1<<PS2_BIT_PARITY)				; Also, calculate parity as we go
			xorwf	PS2_FLAGS,F
			nop

			call	PS2_WaitClockHigh				; Wait for the device to bring Clock high.
			btfsc	PS2_FLAGS, PS2_BIT_ERROR
			goto	PS2_WB_fail

			decfsz	PS2_COUNT,F						; Loop until all bits are sent
			goto	PS2_WB_Loop

			;------ PARITY BIT ------
			call	PS2_WaitClockLow				; Wait for the device to bring the Clock line low.
			btfsc	PS2_FLAGS, PS2_BIT_ERROR
			goto	PS2_WB_fail

			btfss	PS2_FLAGS,PS2_BIT_PARITY		; Put the parity bit on the Data line
			goto	$+2
			goto	$+3
			bcf		PS2_DAT_PORT, PS2_DAT_PIN
			goto	$+2
			bsf		PS2_DAT_PORT, PS2_DAT_PIN
			nop
		
			call	PS2_WaitClockHigh				; Wait for the device to bring Clock high.
			btfsc	PS2_FLAGS, PS2_BIT_ERROR
			goto	PS2_WB_fail

			;------ RELEASE DATA LINE -----
			banksel	PS2_DAT_TRIS					; Release the Data line by setting..
			bsf		PS2_DAT_TRIS, PS2_DAT_PIN		; ..the PS2 Data pin for input
			banksel	PS2_DAT_PORT

			;------ STOP BIT ------
			call	PS2_WaitClockLow				; Wait for the device to bring Clock low.
			btfsc	PS2_FLAGS, PS2_BIT_ERROR
			goto	PS2_WB_fail

			call	PS2_WaitClockHigh				; Wait for the device to bring Clock high.
			btfsc	PS2_FLAGS, PS2_BIT_ERROR
			goto	PS2_WB_fail

			;------ WAIT FOR ACK -----
			call	PS2_WaitDataLow					; Wait for the device to bring Data low.
			btfsc	PS2_FLAGS, PS2_BIT_ERROR
			goto	PS2_WB_fail

			call	PS2_WaitClockLow				; Wait for the device to bring Clock low
			btfsc	PS2_FLAGS, PS2_BIT_ERROR
			goto	PS2_WB_fail

			call	PS2_WaitClockHigh				; Wait for the device to bring Clock high.
			btfsc	PS2_FLAGS, PS2_BIT_ERROR
			goto	PS2_WB_fail

			call	PS2_WaitDataHigh				; Wait for the device to release Data and Clock.
			btfsc	PS2_FLAGS, PS2_BIT_ERROR
			goto	PS2_WB_fail

			goto	PS2_WB_exit

PS2_WB_fail
			bsf		PS2_FLAGS, PS2_BIT_ERROR
			movlw	PS2_ERR_WRITE_FAIL
			movwf	PS2_RESULT
PS2_WB_exit
			return

		;---------------------------------------------
		;	PS2: Read Byte
		;---------------------------------------------
PS2_ReadByte
			clrf	PS2_STEP

			
			;----- START BIT -----
			call	PS2_WaitClockLow				; Loop here until clock goes low
			btfsc	PS2_FLAGS, PS2_BIT_ERROR
			goto	PS2_RB_fail
			incf	PS2_STEP,F						; Increment debug progress field
		
			bcf		STATUS, C						; Read bit into Carry
			btfsc	PS2_DAT_PORT, PS2_DAT_PIN
			bsf		STATUS, C

			; MUST BE LOW, OTHERWISE EXIT
			; ---> TODO <---

			incf	PS2_STEP,F						; Increment debug progress field
			
			call	PS2_WaitClockHigh				; Loop here until clock goes high
			btfsc	PS2_FLAGS, PS2_BIT_ERROR
			goto	PS2_RB_fail
			incf	PS2_STEP,F						; Increment debug progress field
			
			movlw	8								; Initialize bit counter and data
			movwf	PS2_COUNT
			clrf	PS2_DATA
			
			;----- DATA BITS -----
PS2_RB_ReadBits
			call	PS2_WaitClockLow				; Loop here until clock goes low
			btfsc	PS2_FLAGS, PS2_BIT_ERROR
			goto	PS2_RB_fail
			incf	PS2_STEP,F						; Increment debug progress field

			bcf		STATUS, C						; Read bit into Carry
			btfsc	PS2_DAT_PORT, PS2_DAT_PIN
			bsf		STATUS, C
			rrf		PS2_DATA,F						; Shift into byte

			call	PS2_WaitClockHigh				; Loop here until clock goes high
			btfsc	PS2_FLAGS, PS2_BIT_ERROR
			goto	PS2_RB_fail
			incf	PS2_STEP,F						; Increment debug progress field

			decfsz	PS2_COUNT,F						; Loop until done
			goto	PS2_RB_ReadBits

			;----- PARITY BIT -----
			call	PS2_WaitClockLow				; Read Parity Bit
			btfsc	PS2_FLAGS, PS2_BIT_ERROR
			goto	PS2_RB_fail
			incf	PS2_STEP,F						; Increment debug progress field

			bcf		STATUS, C						; Clock is now low. Read bit into Carry
			btfsc	PS2_DAT_PORT, PS2_DAT_PIN
			bsf		STATUS, C
			
			call	PS2_WaitClockHigh				; Loop here until clock goes high
			btfsc	PS2_FLAGS, PS2_BIT_ERROR
			goto	PS2_RB_fail
			incf	PS2_STEP,F						; Increment debug progress field

			;----- STOP BIT -----
			call	PS2_WaitClockLow				; Read Stop Bit. Loop here until clock goes low
			btfsc	PS2_FLAGS, PS2_BIT_ERROR
			goto	PS2_RB_fail
			incf	PS2_STEP,F						; Increment debug progress field

			bcf		STATUS, C						; Clock is now low. Read bit into Carry
			btfsc	PS2_DAT_PORT, PS2_DAT_PIN
			bsf		STATUS, C

			; MUST BE HIGH, OTHERWISE EXIT
			; ---> TODO <---

			call	PS2_WaitClockHigh				; Loop here until clock goes high
			btfsc	PS2_FLAGS, PS2_BIT_ERROR
			goto	PS2_RB_fail
			incf	PS2_STEP,F						; Increment debug progress field
			goto	PS2_RB_exit						; Successful exit

PS2_RB_fail
			bsf		PS2_FLAGS, PS2_BIT_ERROR		; Set the error bit and..
			movlw	PS2_ERR_READ_FAIL				; ..the error value
			movwf	PS2_RESULT
PS2_RB_exit
			return

		;---------------------------------------------
		;	PS2: Wait Clock Low
		;
		;	This routine waits for for the clock line
		;	to go low. It clears the error bit on 
		;	entry. If the clock line does not go high
		; 	after a certain amount of time, it sets
		;	the error bit and returns.
		;---------------------------------------------
PS2_WaitClockLow
			bcf		PS2_FLAGS, PS2_BIT_ERROR		; Clear the error bit
			movlw	PS2_LEVEL_CHANGE_TIMEOUT
			movwf	TEMP
			btfss	PS2_CLK_PORT, PS2_CLK_PIN
			goto	PS2_WCL_exit
			nop
			nop
			decfsz	TEMP,F
			goto	$-5

			bsf		PS2_FLAGS, PS2_BIT_ERROR		; Set the error bit
PS2_WCL_exit		
			return

		;---------------------------------------------
		;	PS2: Wait Clock High
		;
		;	Identical to PS_WaitClockLow, but waiting
		;	for clock to go high
		;---------------------------------------------
PS2_WaitClockHigh
			bcf		PS2_FLAGS, PS2_BIT_ERROR		; Clear the error bit
			movlw	PS2_LEVEL_CHANGE_TIMEOUT
			movwf	TEMP
			btfsc	PS2_CLK_PORT, PS2_CLK_PIN
			goto	$+6
			nop
			nop
			decfsz	TEMP,F
			goto	$-5
			bsf		PS2_FLAGS, PS2_BIT_ERROR		; Set the error bit
			return

		;---------------------------------------------
		;	PS2: Wait Data Low
		;
		;	Identical to PS_WaitClockLow, but waiting
		;	for data to go low
		;---------------------------------------------
PS2_WaitDataLow
			bcf		PS2_FLAGS, PS2_BIT_ERROR		; Clear the error bit
			movlw	PS2_LEVEL_CHANGE_TIMEOUT		; Wait for the device to bring Data low
			movwf	TEMP
			btfss	PS2_DAT_PORT, PS2_DAT_PIN
			goto	$+6
			nop
			nop
			decfsz	TEMP,F
			goto	$-5
			bsf		PS2_FLAGS, PS2_BIT_ERROR		; Set the error bit
			return

		;---------------------------------------------
		;	PS2: Wait Data High
		;
		;	Identical to PS_WaitClockHigh, but waiting
		;	for data to go high
		;---------------------------------------------
PS2_WaitDataHigh
			bcf		PS2_FLAGS, PS2_BIT_ERROR		; Clear the error bit
			movlw	PS2_LEVEL_CHANGE_TIMEOUT		; Set the timeout counter
			movwf	TEMP
			btfsc	PS2_DAT_PORT, PS2_DAT_PIN		; Is the line still low?
			goto	$+6								; No. Goto to exit
			nop										; Yes. Delay a few usecs.
			nop
			decfsz	TEMP,F							; Did loop end?
			goto	$-5								; No. Go back to test for line level
			bsf		PS2_FLAGS, PS2_BIT_ERROR		; Yes. Set the error bit because of timeout.
			return


;============================================================================================
;=====
;=====		Full-frame image load routines
;=====
;============================================================================================


		;---------------------------------------------
		;	LoadCheckerboard
		;	
		;---------------------------------------------
LoadCheckerboard
			; Checkerboard
			movlw	0x55		; Col 1		
			movwf	FRAME
			movwf	FRAME+1
			movlw	0xaa		; Col 2
			movwf	FRAME+2
			movwf	FRAME+3
			movlw	0x55		; Col 3
			movwf	FRAME+4
			movwf	FRAME+5
			movlw	0xaa		; Col 4
			movwf	FRAME+6
			movwf	FRAME+7
			movlw	0x55		; Col 5
			movwf	FRAME+8
			movwf	FRAME+9
			movlw	0xaa		; Col 6
			movwf	FRAME+10
			movwf	FRAME+11
			movlw	0x55		; Col 7
			movwf	FRAME+12
			movwf	FRAME+13
			movlw	0xaa		; Col 8
			movwf	FRAME+14
			movwf	FRAME+15
			movlw	0x55		; Col 9
			movwf	FRAME+16
			movwf	FRAME+17
			return

		;---------------------------------------------
		;	LoadConcentricRects
		;	
		;---------------------------------------------
LoadConcentricRects

			; Concentric Rectangles
			movlw	0xff		; Col 1		
			movwf	FRAME
			movlw	0xff
			movwf	FRAME+1
			movlw	0x01		; Col 2
			movwf	FRAME+2
			movlw	0x04
			movwf	FRAME+3
			movlw	0xfd		; Col 3
			movwf	FRAME+4
			movlw	0x05
			movwf	FRAME+5
			movlw	0x05		; Col 4
			movwf	FRAME+6
			movlw	0x55
			movwf	FRAME+7
			movlw	0x75		; Col 5
			movwf	FRAME+8
			movlw	0x05
			movwf	FRAME+9
			movlw	0x05		; Col 6
			movwf	FRAME+10
			movlw	0x55
			movwf	FRAME+11
			movlw	0xfd		; Col 7
			movwf	FRAME+12
			movlw	0x05
			movwf	FRAME+13
			movlw	0x01		; Col 8		
			movwf	FRAME+14
			movlw	0x04
			movwf	FRAME+15
			movlw	0xff		; Col 9	
			movwf	FRAME+16
			movlw	0x07
			movwf	FRAME+17
			return



		;---------------------------------------------
		;	LoadStar
		;	
		;---------------------------------------------
LoadStar
			movlw	0x23		; Col 1		
			movwf	FRAME
			movlw	0x06
			movwf	FRAME+1
			movlw	0xf9		; Col 2
			movwf	FRAME+2
			movlw	0x04
			movwf	FRAME+3
			movlw	0x8d		; Col 3
			movwf	FRAME+4
			movlw	0x05
			movwf	FRAME+5
			movlw	0x24		; Col 4
			movwf	FRAME+6
			movlw	0x01
			movwf	FRAME+7
			movlw	0x76		; Col 5
			movwf	FRAME+8
			movlw	0x03
			movwf	FRAME+9
			movlw	0x24		; Col 6
			movwf	FRAME+10
			movlw	0x01
			movwf	FRAME+11
			movlw	0x8d		; Col 7
			movwf	FRAME+12
			movlw	0x05
			movwf	FRAME+13
			movlw	0xf9		; Col 8		
			movwf	FRAME+14
			movlw	0x04
			movwf	FRAME+15
			movlw	0x23		; Col 9	
			movwf	FRAME+16
			movlw	0x06
			movwf	FRAME+17
			return

			end
